unit word32;

interface

uses
 windows,sysutils,worddef;

Const
  MAX_ARGS = 34;  { MaxArgs based on largest dialog }

Type
  TParmKind = (pkNone, pkShort, pkLong, pkDouble, pkString, pkWrite, pkRead );
  TParmFlag = Set of TParmKind;

  TArrayDef = record
    cArrayDimensions : Integer;
    arrayDimensions : Array[0..0] of Byte;
  end;
  PArrayDef = ^TArrayDef;

  PDoubleArray = Pointer;
  PStringArray = ^PChar;

  PParm = ^TParm;
  TParm = record
    Value : record
    case Integer of
      0 : (AsShort  : Integer);
      1 : (AsLong   : LongInt);
      2 : (AsDouble : Double);
      3 : (AsString : PChar);
      4 : (AsArray  : PArrayDef;
           Ptr : record
           case Boolean of
             FALSE : (DoubleArray : PDoubleArray);
             TRUE  : (StringArray : PStringArray);
           end;
          );
    end;
    Size  : Word;
    Data  : record
     Case Byte of
      0 : (Kind  : TParmKind);
      1 : (Flag  : TParmFlag);
      2 : (Value : word); { so, Field is 2 bytes long }
    end;
    FldID : Word;
   end;

  TParmEditor = class
   Private
    Parm:^TParm;
    Procedure SetDataKind(Value:TParmKind);
    Procedure SetAsShort (Value:Integer);
    Procedure SetAsLong  (Value:LongInt);
    Procedure SetAsDouble(Value:Double);
    Procedure SetAsString(Value:String);
   Public
    Property AsShort:Integer Write SetAsShort;
    Property AsLong:LongInt  Write SetAsLong;
    Property AsDouble:double Write SetAsDouble;
    Property AsString:String Write SetAsString;
    Procedure SetFieldID(Value:integer);
    Procedure GetFieldID(Value:integer);
  end;

  PParmList = ^TParmList;
  TParmList = Array[0..MAX_ARGS-1] of TParm;

  TWordMsg = record
   ParmCount : Integer;
   Parms     : PParmList;
   Return    : TParm;
  end;

  TDlgOption = (doDefaults,    { GetCurValues }
                doDialog,      { display dialog }
                doAction,      { execute dialog }
                doDlgAction);  { display & exec }

 TWordBasic=class
  DocId:integer;
  WordMsg:TWordMsg;
  fParmCount:integer;
  fParmSize :integer;
  ParmEditor:TParmEditor;
  Procedure SetParmCount(Value:integer);
  Procedure SetParmSize(Value:integer);
  Function  GetParm(Index:integer):TParmEditor;
  Constructor Create(ADocId:Integer);
  Destructor  Destroy; override;
// Execute
  Function Execute(CmdId:Integer; ReturnKind:TParmKind):integer;
  Function ExecuteDlg(CmdId:Integer; ReturnKind:TParmKind; DlgOption:TDlgOption):integer;
// Menu
  Procedure RegisterFunction(Name,Title:string);
  Procedure AddMenu(Caption:string; Index,Flag:integer);
  Procedure AddMenuItem(Menu,Caption,Action:string; Index,Flag:integer);
// Text
  Procedure Insert(Text:string);
// Properties
  Property ParmCount:integer read fParmCount write SetParmCount;
  Property ParmSize :integer read fParmSize  Write SetParmSize;
  Property Parms[Index:Integer]:TParmEditor read GetParm;
 end;

implementation

Var
 wdCommandDispatch:function(Cmd,Dlg,Args:integer; Parms:PParmList;Return:PParm):integer; stdcall;

{ TParmEditor }

Procedure TParmEditor.SetDataKind;
 begin
  Parm^.Data.Value:=Ord(Value);
  Parm^.Size:=0;
  Parm^.FldID:=0;
 end;

Procedure TParmEditor.SetAsShort;
 begin
  SetDataKind(pkShort);
  Parm^.Value.AsShort:=Value;
 end;

Procedure TParmEditor.SetAsLong;
 begin
  SetDataKind(pkLong);
  Parm^.Value.AsLong:=Value;
 end;

Procedure TParmEditor.SetAsDouble;
 begin
  SetDataKind(pkDouble);
  Parm^.Value.AsDouble:=Value;
 end;

Procedure TParmEditor.SetAsString;
 begin
  SetDataKind(pkString);
  Parm^.Value.AsString:=pchar(Value);
  Parm^.Size:=Length(Value);
 end;

Procedure TParmEditor.SetFieldID;
 begin
  Include(Parm^.Data.Flag,pkWrite);
  Parm^.FldID:=Value;
 end;
 
Procedure TParmEditor.GetFieldID;
 begin
  Include(Parm^.Data.Flag,pkRead);
  Parm^.FldID:=Value;
 end;



Constructor TWordBasic.Create(ADocId:integer);
 begin
  DocId:=ADocId;
  fParmCount:=0;
  fParmSize :=0;
  ParmEditor:=TParmEditor.Create;
 end;

Destructor TWordBasic.Destroy;
 begin
  ParmSize:=0;
  ParmEditor.Free;
  inherited Destroy;
 end;

Function TWordBasic.Execute(CmdId:Integer; ReturnKind:TParmKind):integer;
 begin
  With WordMsg do begin
   if ReturnKind=pkNone then // Word return error #5007 is @Return<>nil  
    Result:=wdCommandDispatch(CmdID,0,ParmCount,Parms,nil)
   else begin
    Return.Data.Value:=ord(ReturnKind);
    Result:=wdCommandDispatch(CmdID,0,ParmCount,Parms,@Return);
   end;
   if Result>0 then MessageBox(0,pchar('Word error #'+IntToStr(Result)),'XWord-32',0);
  end;
 end;

Function TWordBasic.ExecuteDlg(CmdId:Integer; ReturnKind:TParmKind; DlgOption:TDlgOption):integer;
 begin
  With WordMsg do begin
   if ReturnKind=pkNone then
    Result:=wdCommandDispatch(CmdID,integer(DlgOption),ParmCount,Parms,nil)
   else begin
    Return.Data.Value:=ord(ReturnKind);
    Result:=wdCommandDispatch(CmdID,integer(DlgOption),ParmCount,Parms,@Return);
   end;
  end;
 end;

Procedure TWordBasic.SetParmCount(Value:integer);
 begin
  WordMsg.ParmCount:=Value;
  If fParmCount<Value then begin
   ParmSize:=Value*SizeOf(TParm);
   fParmCount:=Value;
  end;
 end;

Procedure TWordBasic.SetParmSize(Value:integer);
 Var
  NewParms:pointer;
 begin
  If Value>0 then GetMem(NewParms,Value) else NewParms:=nil;
  if fParmSize>0 then begin
   if Value>0 then Move(WordMsg.Parms^,NewParms^,Value);
   FreeMem(WordMsg.Parms,fParmSize);
  end;
  fParmSize:=Value;
  WordMsg.Parms:=NewParms;
 end;

Function TWordBasic.GetParm(Index:Integer):TParmEditor;
 begin
  If Index>=ParmCount then ParmCount:=Index+1; // alloc if needed
  ParmEditor.Parm:=@WordMsg.Parms^[Index];     // select Parm
  Result:=ParmEditor;
 end;

// Menu
Procedure TWordBasic.RegisterFunction(Name,Title:string);
 begin
  ParmCount:=3;
  Parms[0].AsShort:=DocId;
  Parms[1].AsString:=Name;
  Parms[2].AsString:=Title;
  Execute(wdAddCommand,pkNone);
 end;

Procedure TWordBasic.AddMenu(Caption:string; Index,Flag:integer);
 begin
  ParmCount:=5;
  Parms[0].AsString:=Caption;   Parms[0].SetFieldId(fidMenuText);
  Parms[1].AsShort:=Index;      Parms[1].SetFieldId(fidPosition);
  Parms[2].AsShort:=1;          Parms[2].SetFieldId(fidAdd);
  Parms[3].AsShort:=Flag;       Parms[3].SetFieldId(fidMenuType);
  Parms[4].AsShort:=DocId;      Parms[4].SetFieldId(fidContext);
  ExecuteDlg(wdToolsCustomizeMenuBar,pkShort,doAction);
 end;

Procedure TWordBasic.AddMenuItem(Menu,Caption,Action:string; Index,Flag:integer);
 begin
  ParmCount:=8;
  Parms[0].AsShort:=DocID;           Parms[0].SetFieldId(fidContext);
  Parms[1].AsString:=Menu;           Parms[1].SetFieldId(fidMenu);
  Parms[2].AsString:=Action;         Parms[2].SetFieldId(fidName);
  Parms[3].AsString:=Caption;        Parms[3].SetFieldId(fidMenuText);
  Parms[4].AsShort:=Length(Caption); Parms[4].SetFieldId(fidPosition);
  Parms[5].AsShort:=Flag;            Parms[5].SetFieldId(fidMenuType);
  Parms[6].AsShort:=1;               Parms[6].SetFieldId(fidCategory);
  Parms[7].AsShort:=1;               Parms[7].SetFieldId(fidAdd);
  ExecuteDlg(wdToolsCustomizeMenus,pkShort,doAction);
 end;

// Text
Procedure TWordBasic.Insert(Text:string);
 begin
  ParmCount:=1;
  Parms[0].AsString:=Text;
  Execute(wdInsert,pkNone);
 end;


initialization
 @wdCommandDispatch:=GetProcAddress(GetModuleHandle(nil),'wdCommandDispatch');
 if @wdCommandDispatch=nil then halt;
end.
